/**
 * \file main.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date December 2002
 * \brief The code for instruction-level-reduction.
 *
 * This file contains all of the code for the
 * instruction-level-reduction reference handler, which converts the
 * binary output of the Laplace raw reference trace a
 * difference-encoded text format.
 */



// ===================================================================
#include <math.h> // For pow()
#include <stdio.h>
#include <stdlib.h>
// ===================================================================



// =================================================================
// TYPES
// Definitions of types for each field in the raw reference trace.
// This is taken from tracer.cc in Laplace itself.

/** \brief The type for the tag that indicates the type of a
 *         reference. */
typedef char tag_t;

/** \brief The type for the cycle counter timestamp. */
typedef unsigned long long int timestamp_t;

/** \brief The type for length values. */
typedef unsigned char length_t;

/** \brief The type for the address space identifier, a.k.a. the page
 *         number holding the base level of the page table. */
typedef unsigned long int address_space_ID_t;

/** \brief The type for virtual addresses. */
typedef unsigned long int address_t;

/** \brief A record structure in which a reference record will be
 *         passed and stored. */
struct reference_record_s {

  tag_t tag;
  timestamp_t timestamp;
  length_t length;
  address_space_ID_t address_space_ID;
  address_t virtual_address;

};
// =================================================================



// =================================================================
// GLOBALS and CONSTANTS

/** \brief The tag to mark the end of the trace. */
const tag_t TAG_END_OF_TRACE = 'e';
// ===================================================================



// ===================================================================
/**
 * \brief The actual engine for handling references.
 *
 * Read every record from the raw reference trace (via stdin), perform
 * difference encoding on the values (as described in the README
 * file), and emit a text representation.
 **/

void
process_trace () {

  // Allocate a single reference record into which the binary source
  // will be read, one record at a time.
  reference_record_s binary_record;

  // The previous timestamp.
  timestamp_t previous_timestamp = 0;

  // The previous virtual address space ID (tracked without regard to
  // reference type).
  address_space_ID_t previous_address_space_ID = 0;

  // An array storing the previous virtual addresses of each type,
  // initialized.  Note that for efficiency, we allocate an array that
  // is large enough so that there is an entry for each possible tag
  // value.
  const unsigned int bits_per_byte = 8;
  const unsigned int number_possible_tags =
    (unsigned int)pow(2.0, (double)(sizeof(tag_t) * bits_per_byte));
  address_t previous_address[number_possible_tags];
  for (unsigned int i = 0; i < number_possible_tags; i++) {
    previous_address[i] = 0;
  }

  // A record number counter, used for debugging purposes.
  unsigned long long int record_number = 1;

  // Read binary records until the end of the trace has been reached.
  do {

    // Attempt to read the next record.
    size_t read_result = fread(&binary_record,
			       sizeof(binary_record),
			       1,
			       stdin);

    // Ensure that the read succeeded; otherwise, bail out.
    if (read_result != 1) {

      if (feof(stdin)) {
	fprintf(stderr, "Premature EOF\n");
      } else {
	fprintf(stderr, "Failed in reading binary record\n");
      }

      exit(1);

    }

    // Have we reached the end of the trace?
    if (binary_record.tag == TAG_END_OF_TRACE) {

      // Yes, so emit the ending tag.
      printf("%c\n", binary_record.tag);

    } else {

      // Calculate the timestamp difference and adopt the new
      // timestamp.
      timestamp_t timestamp_difference =
	binary_record.timestamp - previous_timestamp;
      previous_timestamp = binary_record.timestamp;

      // Calculate the address space ID to emit.  Is it the same as
      // the previous address space ID, emit zero; otherwise, emit the
      // changed address space ID.
      address_space_ID_t address_space_ID = 0;
      if (binary_record.address_space_ID !=
	  previous_address_space_ID) {
	address_space_ID = binary_record.address_space_ID;
	previous_address_space_ID = address_space_ID;
      }

      // Calculate the virtual address difference based on the
      // reference type.
      char address_difference_sign;
      address_t address_difference_magnitude;
      if (binary_record.virtual_address <
	  previous_address[binary_record.tag]) {

	address_difference_sign = '-';
	address_difference_magnitude =
	  previous_address[binary_record.tag] -
	  binary_record.virtual_address;

      } else {

	address_difference_sign = '+';
	address_difference_magnitude =
	  binary_record.virtual_address - 
	  previous_address[binary_record.tag];

      }
      previous_address[binary_record.tag] =
	binary_record.virtual_address;

      printf("%c %qx %hx %lx %c %lx\n",
	     binary_record.tag,
	     timestamp_difference,
	     binary_record.length,
	     address_space_ID,
	     address_difference_sign,
	     address_difference_magnitude);

      record_number++;

    }

  } while (binary_record.tag != TAG_END_OF_TRACE);

} // process_trace
// ===================================================================



// ===================================================================
/**
 * \brief The entry point.
 * 
 * \param argc The number of arguments given at the command line.
 * \param argv The vector of argument strings passed at the command
 *             line.
 * \return An execution result code.
 * 
 * Parse the command line arguments.  For this utility, there should
 * be none, as input is taken from stdin, and output is sent to
 * stdout.
 **/

int
main (const unsigned int argc, const char** const argv) {

  // Were the correct number of arguments passed?
  if (argc != 1) {

    // No.  Emit usage and exit.
    fprintf(stderr, "Usage: %s <NO ARGUMENTS>\n", argv[0]);
    exit(1);

  }

  // Perform the reference handling.
  process_trace();

  // Exit successfully.
  return 0;

} // main
// ===================================================================
